import threading
from importlib.metadata import entry_points

from ed_core import core, cache
from ed_core.config import load_config, apply_config

VERSION = "2023.5.11"
BACKENDS = {}
print(cache.CACHE)


def run(standalone: bool = True):
    try:
        print("El Decko Core running.")
        load_config()
        discovered_backends = entry_points(group='eldecko.backend')
        for edb in discovered_backends:
            edb_id = edb.value.split(":")[0]
            edb_function = edb.load()
            edb_name: str = edb.name
            if edb_name.lower() == "init":
                edb_function()
            else:
                if edb_id not in BACKENDS:
                    BACKENDS[edb_id] = {}
                BACKENDS[edb_id][edb_name] = edb_function

        core.initialize(BACKENDS)
        if standalone:
            for t in threading.enumerate():
                try:
                    t.join()
                except RuntimeError:
                    pass
    except KeyboardInterrupt:
        stop_core()


def stop_core():
    print("Stopping El Decko Core")
    core.shutdown()
    __stop_backends()


def __stop_backends():
    for backend in BACKENDS:
        print(backend)
        shutdown = BACKENDS[backend]["stop"]
        shutdown()


def backends():
    return BACKENDS



from StreamDeck.Devices.StreamDeck import StreamDeck
from StreamDeck.DeviceManager import DeviceManager

from ed_core import config

backends = []
__stream_decks = DeviceManager().enumerate()


def initialize(edbs):
    global backends
    backends = edbs
    for index, deck in enumerate(__stream_decks):
        if not deck.is_visual():
            continue

        deck.open()
        deck.reset()

        print("Opened '{}' device (serial number: '{}', fw: '{}')".format(
            deck.deck_type(), deck.get_serial_number(), deck.get_firmware_version()
        ))

        config.apply_config(deck)
        deck.set_key_callback(__key_change_callback)


def shutdown():
    for index, deck in enumerate(__stream_decks):
        if not deck.is_visual():
            continue
        deck.reset()
        deck.close()


def set_brightness(deck, brightness: int):
    deck.set_brightness(brightness)


def get_key_config(deck_serial: str, key_num: int):
    config.load_config()
    return config.DECK_CFG[deck_serial]["key_config"][str(key_num)]


def __key_change_callback(deck: StreamDeck, key, state):
    if state:
        cfg = config.DECK_CFG[deck.get_serial_number()]
        edb_id: str = cfg["key_config"][str(key)]["backend"]
        edb_event: str = cfg["key_config"][str(key)]["event"]
        edb_params: dict = cfg["key_config"][str(key)]["event_parameters"]
        if edb_id:
            fire = backends[edb_id]["fire"]
            fire(edb_event, edb_params)



from StreamDeck.DeviceManager import DeviceManager

CACHE = list()


# Builds a cache of all connected Stream Decks for interaction ith UIs as someone can not interact with a
# Stream Deck if it is already opened. Hence, if the core is running.
def build_cache():
    for index, deck in enumerate(DeviceManager().enumerate()):
        properties = {}
        deck.open()
        properties["serial"] = deck.get_serial_number()
        properties["key_count"] = deck.key_count()
        properties["key_layout"] = deck.key_layout()
        properties["key_image_format"] = deck.key_image_format()
        properties["deck_type"] = deck.deck_type()
        properties["firmware_version"] = deck.get_firmware_version()
        properties["id"] = deck.id()
        properties["product_id"] = deck.product_id()
        properties["vendor_id"] = deck.vendor_id()
        CACHE.append(properties)
        deck.close()


build_cache()



import os
import json
import shutil

from xdg import (
    xdg_cache_home,
    xdg_config_dirs,
    xdg_config_home,
    xdg_data_dirs,
    xdg_data_home,
    xdg_runtime_dir,
    xdg_state_home,
)

__app_name = "eldecko"
CONFIG_ROOT: str = ""
DATA_ROOT: str = ""
if os.name == "nt":
    CONFIG_ROOT = os.getenv('LOCALAPPDATA') + "/" + __app_name + "/"
    DATA_ROOT = os.getenv('LOCALAPPDATA') + "/" + __app_name + "/"
elif os.name == "posix":
    CONFIG_ROOT = str(xdg_config_home()) + "/" + __app_name + "/"
    DATA_ROOT = str(xdg_data_home()) + "/" + __app_name + "/"
else:
    CONFIG_ROOT = ""
    DATA_ROOT = ""

CONFIG_ROOT_BACKEND = CONFIG_ROOT + "backends/"
ASSETS_ROOT = DATA_ROOT + "assets/"

if not os.path.exists(ASSETS_ROOT):
    os.makedirs(ASSETS_ROOT)
if not os.path.exists(CONFIG_ROOT):
    os.makedirs(CONFIG_ROOT)


def load_config(root_path: str, file_name: str):
    if not os.path.exists(root_path):
        os.makedirs(root_path)
    try:
        with open(root_path + file_name) as input_file:
            data = json.load(input_file)
            return data
    except json.decoder.JSONDecodeError as e:
        print(e)


def store_config(root_path: str, file_name: str, data: dict):
    if not os.path.exists(root_path):
        os.makedirs(root_path)
    with open(root_path + file_name, "w+", encoding="utf-8") as outfile:
        json.dump(data, outfile, ensure_ascii=False, indent=2)


def store_image(source: str):
    file_name = os.path.basename(source).split('/')[-1]
    shutil.copyfile(source, ASSETS_ROOT + file_name)


def config_exists(root_path: str, file_name: str):
    return os.path.isfile(root_path + file_name)



import json
import os

from StreamDeck.DeviceManager import DeviceManager
from StreamDeck.Devices.StreamDeck import StreamDeck
from StreamDeck.ImageHelpers import PILHelper
from PIL import Image, ImageDraw, ImageFont

from ed_core import dyn_data

CONFIG_FILE: str = "streamdeck.json"
DECK_CFG: dict = {}


def load_config():
    if not dyn_data.config_exists(dyn_data.CONFIG_ROOT, CONFIG_FILE):
        __create_default_config()
    global DECK_CFG
    if not DECK_CFG:
        DECK_CFG = dyn_data.load_config(dyn_data.CONFIG_ROOT, CONFIG_FILE)


def get_config():
    return DECK_CFG


def apply_config(deck: StreamDeck):
    serial = deck.get_serial_number()
    deck.set_brightness(DECK_CFG[serial]["brightness"])
    for i in range(0, deck.key_count()):
        image_path: str = DECK_CFG[deck.get_serial_number()]["key_config"][str(i)]["image_idle"]
        label: str = DECK_CFG[deck.get_serial_number()]["key_config"][str(i)]["label"]
        if image_path and image_path.startswith(dyn_data.ASSETS_ROOT) and os.path.isfile(image_path):
            icon = Image.open(image_path)
            image = PILHelper.create_scaled_image(deck, icon)
            if label != "":
                image = PILHelper.create_scaled_image(deck, icon, margins=[0, 0, 20, 0])
                draw = ImageDraw.Draw(image)
                font = ImageFont.truetype("NotoSans-Regular.ttf", 12)  # ToDo: Replace with actual ttf from assets
                draw.text((image.width / 2, image.height - 5), text=label, font=font, anchor="ms", fill="white")
            native_image = PILHelper.to_native_format(deck, image)
            deck.set_key_image(i, native_image)
        else:
            print("Invalid icon path: {} for Deck: {} and key: {}".format(image_path, deck.get_serial_number(), i))
            print("Image must be located at: {}".format(dyn_data.ASSETS_ROOT))


def __create_key_image(deck, image_file, font_style, label):
    pass


def __create_default_config():
    stream_decks = DeviceManager().enumerate()
    for index, deck in enumerate(stream_decks):
        if not deck.is_visual():
            continue

        deck.open()
        deck.reset()

        serial = deck.get_serial_number()

        key_config: dict = {}
        for i in range(0, deck.key_count()):
            key_config[str(i)] = {
                "backend": None,
                "event": None,
                "event_parameters": {

                },
                "image_idle": None,
                "image_pressed": None,
                "label": ""
            }

        DECK_CFG[serial] = dict(brightness=30, key_config=key_config)
        deck.close()
        dyn_data.store_config(dyn_data.CONFIG_ROOT, CONFIG_FILE, DECK_CFG)
